import logging
from typing import ClassVar

import torch
import torch.nn.functional as F
from meds_torchdata import MEDSTorchBatch
from omegaconf import DictConfig
from transformers import (
    AutoConfig,
    AutoModelForCausalLM,
    GenerationConfig,
    GPTNeoXConfig,
    GPTNeoXForCausalLM,
)
from transformers.modeling_outputs import CausalLMOutputWithPast

logger = logging.getLogger(__name__)

try:
    import flash_attn  # noqa: F401

    HAS_FLASH_ATTN = True
    logger.info("FlashAttention is available.")
except ImportError:
    HAS_FLASH_ATTN = False


def _val(tensor: torch.Tensor) -> int | bool | float:
    """Returns the value of a scalar-tensor as a Python scalar.

    Examples:
        >>> _val(torch.tensor(1))
        1
        >>> _val(torch.tensor(1.0))
        1.0
        >>> _val(torch.tensor(False))
        False
    """
    return tensor.detach().cpu().item()


class Model(torch.nn.Module):
    """A basic GPT-NeoX like model for pre-training an autoregressive, "everything-is-code" model.

    This model is a wrapper around the Hugging Face GPTNeoXForCausalLM model to run it over MEDS-TorchData
    batches.

    Args:
        gpt_kwargs: A dictionary of keyword arguments to pass to the GPTNeoXConfig constructor. These can
            include 'max_position_embeddings', 'vocab_size', 'hidden_size', etc.

    Raises:
        ValueError: If the gpt_kwargs contains a key that is not supported.

    Examples:
        >>> _ = torch.manual_seed(0)
        >>> model = Model({
        ...     "num_hidden_layers": 2,
        ...     "num_attention_heads": 2,
        ...     "hidden_size": 4,
        ...     "max_position_embeddings": 10,
        ...     "vocab_size": dataset_config.vocab_size,
        ... }, precision="16-true")

    Once created, we can use the simple helpers `max_seq_len` and `vocab_size` to access the corresponding
    elements of the model config:

        >>> model.max_seq_len
        10
        >>> model.vocab_size
        39

    We can also run the model on a sample batch of data. This `sample_batch` is defined in our `conftest.py`
    file and is a MEDSTorchBatch object. The only feature of this batch that we use in this model is the
    `code` key.

        >>> sample_batch
        MEDSTorchBatch(code=tensor([[38,  5, 36,  3, 13, 29, 35,  4, 37],
                                    [38,  5, 36,  2, 22, 25, 35,  4, 37]]),
                       ...)

    We run over the batch in the normal way, which internally calls the `forward` method of the model:

        >>> loss, outputs = model(sample_batch)
        >>> print(loss)
        tensor(3.6660, dtype=torch.float16, grad_fn=<NllLoss2DBackward0>)
        >>> print(f"Outputs have keys: {', '.join(outputs.keys())}")
        Outputs have keys: logits, past_key_values
        >>> print(f"Logits shape: {outputs.logits.shape}")
        Logits shape: torch.Size([2, 9, 39])
        >>> print(outputs.logits)
        tensor([[[ 2.0309e-02, ...,  1.9135e-02], ..., [ 1.3763e-02, ...,  1.6571e-02]],
        <BLANKLINE>
                [[ 2.0309e-02, ...,  1.9135e-02], ..., [ 1.4458e-02, ...,  1.6281e-02]]],
               dtype=torch.float16,
               grad_fn=<UnsafeViewBackward0>)

    The models parameters can be accessed in the normal way.

        >>> sample_param_name, sample_param = next(iter(model.named_parameters()))
        >>> print(f"{sample_param_name} ({sample_param.shape}): {sample_param}")
        HF_model.gpt_neox.embed_in.weight (torch.Size([39, 4])): Parameter containing:
        tensor([[-0.0247, -0.0222,  0.0160,  0.0219], ..., [-0.0050, -0.0061, -0.0358,  0.0136]],
               dtype=torch.float16,
               requires_grad=True)

    Let's validate that they have gradients that can be realized via `.backward()` as normal:

        >>> print(f"Sample parameter grad?: {sample_param.grad}")
        Sample parameter grad?: None
        >>> loss.backward()
        >>> print(f"Sample parameter grad?: {sample_param.grad}")
        Sample parameter grad?:
        tensor([[ 0.0000e+00,  0.0000e+00,  0.0000e+00,  0.0000e+00],
                ...,
                [ 1.5251e-02, -6.2347e-02, -3.1921e-02,  7.9102e-02]],
               dtype=torch.float16)

    With a single backward pass, we should not get any infinite gradients:

        >>> for name, param in model.named_parameters():
        ...     if param.grad is not None:
        ...         if not _val(torch.isfinite(param.grad).all()):
        ...             raise ValueError(f"Gradient for {name} is not finite.")

    Model errors are raised if we pass invalid GPT kwargs to the constructor:

        >>> Model({"foobar": 2})
        Traceback (most recent call last):
            ...
        ValueError: Config for HF model gpt-neox does not have attribute foobar
    """

    HF_model_config: GPTNeoXConfig
    HF_model: GPTNeoXForCausalLM
    do_demo: bool
    precision: str

    PRECISION_TO_MODEL_WEIGHTS_DTYPE: ClassVar[dict[str, torch.dtype]] = {
        "32-true": torch.float32,
        "16-true": torch.float16,
        "16-mixed": torch.float32,
        "bf16-true": torch.bfloat16,
        "bf16-mixed": torch.float32,
        "transformer-engine": torch.bfloat16,
    }

    def __init__(self, gpt_kwargs: dict | DictConfig, precision: str = "32-true", do_demo: bool = False):
        super().__init__()

        self.HF_model_config: GPTNeoXConfig = AutoConfig.from_pretrained("EleutherAI/gpt-neox-20b")

        for key, val in gpt_kwargs.items():
            if key == "attention_head_dim":
                continue
            if not hasattr(self.HF_model_config, key):
                raise ValueError(f"Config for HF model gpt-neox does not have attribute {key}")
            setattr(self.HF_model_config, key, val)

        extra_kwargs = {"torch_dtype": self.PRECISION_TO_MODEL_WEIGHTS_DTYPE.get(precision)}

        if HAS_FLASH_ATTN:
            logger.info("Using FlashAttention 2 for the model.")
            extra_kwargs["attn_implementation"] = "flash_attention_2"

        self.HF_model = AutoModelForCausalLM.from_config(self.HF_model_config, **extra_kwargs)

        self.do_demo = do_demo
        if self.do_demo:
            self.forward = self._forward_demo
        else:
            self.forward = self._forward

        self.hparams = {
            "gpt_kwargs": gpt_kwargs,
            "precision": precision,
            "do_demo": do_demo,
        }

    @property
    def max_seq_len(self) -> int:
        """The maximum sequence length of the model."""
        return self.HF_model_config.max_position_embeddings

    @property
    def vocab_size(self) -> int:
        """The vocabulary size of the model."""
        return self.HF_model_config.vocab_size

    def _check_inputs(self, batch: MEDSTorchBatch):
        """Checks the inputs for various validity properties.

        Validity checks:
          - The batch is in "SM" mode (see
            [MEDSTorchBatch](https://meds-torch-data.readthedocs.io/en/latest/api/meds_torchdata/types/#meds_torchdata.types.MEDSTorchBatch)
            for more details).
          - The input sequence length must not exceed the model's maximum sequence length.
          - The input sequence length must not be too short (minimum sequence length is 2).
          - The input must not contain out-of-vocabulary tokens.
          - The input must not contain inf or nan values.
          - The input must not contain only padding tokens.

        Args:
            batch: The input batch of data.

        Raises:
            ValueError: If the input sequence length exceeds the model's maximum sequence length.
            AssertionError: If the input contains out-of-vocabulary tokens or if it contains inf or nan
                values.

        Examples:
            >>> model = Model({
            ...     "num_hidden_layers": 2,
            ...     "num_attention_heads": 2,
            ...     "hidden_size": 4,
            ...     "max_position_embeddings": 3,
            ...     "vocab_size": 10,
            ... })
            >>> batch = Mock(code=torch.LongTensor([[0, 3, 1], [0, 2, 1]]), PAD_INDEX=0, mode="SM")
            >>> model._check_inputs(batch) # no errors
            >>> batch.code = torch.LongTensor([[0, 3, 1], [0, 2, 11]])
            >>> model._check_inputs(batch)
            Traceback (most recent call last):
                ...
            AssertionError: Input sequence contains 1 out-of-vocabulary tokens (max 11 for vocab size 10).
            >>> batch.code = torch.Tensor([[0, 3, 1], [0, 2, float("inf")]])
            >>> model._check_inputs(batch)
            Traceback (most recent call last):
                ...
            AssertionError: Batch code contains inf values.
            >>> batch.code = torch.Tensor([[0, 3, 1], [0, 2, float("nan")]])
            >>> model._check_inputs(batch)
            Traceback (most recent call last):
                ...
            AssertionError: Batch code contains nan values.
            >>> batch.code = torch.LongTensor([[0, 3, 1], [0, 0, 0]])
            >>> model._check_inputs(batch)
            Traceback (most recent call last):
                ...
            AssertionError: 1 samples in the batch have only padding tokens. Batch size: 2, Sequence length: 3
            >>> batch.code = torch.LongTensor([[0, 3, 1, 0], [0, 2, 1, 4]])
            >>> model._check_inputs(batch)
            Traceback (most recent call last):
                ...
            ValueError: Input sequence length 4 exceeds model max sequence length 3.
            >>> batch.code = torch.LongTensor([[1], [2]])
            >>> model._check_inputs(batch)
            Traceback (most recent call last):
                ...
            ValueError: Input sequence length 1 is too short. Minimum sequence length is 2.
            >>> model._check_inputs(Mock(mode="SEM"))
            Traceback (most recent call last):
                ...
            ValueError: Batch mode SEM is not supported.
        """

        code = batch.code

        if batch.mode != "SM":
            raise ValueError(f"Batch mode {batch.mode} is not supported.")

        batch_size, seq_len = code.shape

        if seq_len > self.max_seq_len:
            raise ValueError(
                f"Input sequence length {batch.code.shape[1]} exceeds model max sequence length "
                f"{self.max_seq_len}."
            )
        elif seq_len <= 1:
            raise ValueError(
                f"Input sequence length {batch.code.shape[1]} is too short. Minimum sequence length is 2."
            )

        torch._assert(~torch.isinf(code).any(), "Batch code contains inf values.")
        torch._assert(~torch.isnan(code).any(), "Batch code contains nan values.")

        out_of_vocab = code >= self.vocab_size
        out_of_vocab_msg = (
            f"Input sequence contains {out_of_vocab.sum()} out-of-vocabulary tokens "
            f"(max {batch.code.max()} for vocab size {self.vocab_size})."
        )

        torch._assert(~out_of_vocab.any(), out_of_vocab_msg)

        is_pad = code == batch.PAD_INDEX  # Shape is batch_size x seq_len
        all_samples_pad = is_pad.all(dim=1)  # Shape is batch_size

        all_samples_pad_msg = (
            f"{all_samples_pad.sum()} samples in the batch have only padding tokens. "
            f"Batch size: {code.shape[0]}, Sequence length: {code.shape[1]}"
        )
        torch._assert(~all_samples_pad.any(), all_samples_pad_msg)

    def _check_parameters(self):
        """Logs a warning about the finiteness of any parameters in the model.

        This is only used for advanced debugging. It does not raise an error because when this mode is
        enabled, typically detect anomaly is on in the lightning trainer, and that gives more information
        about these issues than a generic assertion would.

        Validity checks:
            - The parameters are not nan.
            - The parameters are not inf.

        Examples:
            >>> model = Model({
            ...     "num_hidden_layers": 2,
            ...     "num_attention_heads": 2,
            ...     "hidden_size": 4,
            ...     "max_position_embeddings": 3,
            ...     "vocab_size": 10,
            ... })
            >>> model.HF_model.gpt_neox.layers[1].attention.query_key_value.bias.shape
            torch.Size([12])
            >>> model.HF_model.gpt_neox.layers[1].attention.query_key_value.bias = torch.nn.Parameter(
            ...     torch.tensor([float("nan"), 0., float("inf"), 0., 0., 0., 0., 0., 0., 0., 0., 0.])
            ... )
            >>> with print_warnings():
            ...     model._check_parameters()
            Warning: Parameter HF_model.gpt_neox.layers.1.attention.query_key_value.bias contains 1/12 nan
                values.
            Warning: Parameter HF_model.gpt_neox.layers.1.attention.query_key_value.bias contains 1/12 inf
                values.
        """

        for n, p in self.named_parameters():
            num_nan = _val(torch.isnan(p).sum())
            num_inf = _val(torch.isinf(p).sum())

            if num_nan > 0:
                logger.warning(f"Parameter {n} contains {num_nan}/{p.numel()} nan values.")
            if num_inf > 0:
                logger.warning(f"Parameter {n} contains {num_inf}/{p.numel()} inf values.")

    def _check_outputs(self, loss: torch.FloatTensor, outputs: CausalLMOutputWithPast):
        """Logs a warning if the loss is inf or nan.

        This does not raise an error because when this mode is enabled, typically detect anomaly is on in the
        lightning trainer, and that gives more information about these issues than a generic assertion would.

        Validity checks:
            - The loss is not inf or nan.
            - The logits contain inf or nan values.

        Args:
            loss: The loss tensor.

        Examples:
            >>> model = Model({
            ...     "num_hidden_layers": 2,
            ...     "num_attention_heads": 2,
            ...     "hidden_size": 4,
            ...     "max_position_embeddings": 3,
            ...     "vocab_size": 10,
            ... })
            >>> fake_output_valid = Mock(logits=torch.FloatTensor([[0.1, 0.2], [0.3, 0.4]]))
            >>> model._check_outputs(torch.tensor(0.4), fake_output_valid) # no errors
            >>> with print_warnings():
            ...     model._check_outputs(torch.tensor(float("inf")), fake_output_valid)
            ...     model._check_outputs(torch.tensor(float("nan")), fake_output_valid)
            Warning: Loss contains inf values.
            Warning: Loss contains nan values.
            >>> fake_output_inf = Mock(logits=torch.FloatTensor([[float("inf"), 0.2], [0.3, 0.4]]))
            >>> fake_output_nan = Mock(logits=torch.FloatTensor([[0.4, float("nan")], [0.3, float("nan")]]))
            >>> with print_warnings():
            ...     model._check_outputs(torch.tensor(0.4), fake_output_inf)
            ...     model._check_outputs(torch.tensor(0.4), fake_output_nan)
            Warning: Logits contains 1/4 inf values.
            Warning: Logits contains 2/4 nan values.
        """

        if _val(torch.isinf(loss).any()):
            logger.warning("Loss contains inf values.")
        if _val(torch.isnan(loss).any()):
            logger.warning("Loss contains nan values.")

        logits = outputs.logits
        inf_count = _val(torch.isinf(logits).sum())
        if inf_count > 0:
            logger.warning(f"Logits contains {inf_count}/{logits.numel()} inf values.")
        nan_count = _val(torch.isnan(logits).sum())
        if nan_count > 0:
            logger.warning(f"Logits contains {nan_count}/{logits.numel()} nan values.")

    def _hf_inputs(self, batch: MEDSTorchBatch) -> dict[str, torch.Tensor]:
        """Converts the MEDSTorchBatch to a dictionary of inputs for the Hugging Face model.

        HF relevant input keys:
            - input_ids: The input sequence of token IDs. Captured in `batch.code`.
            - attention_mask: A mask to avoid attending to padding tokens. See the
              [documentation](https://huggingface.co/docs/transformers/en/model_doc/gpt_neox#transformers.GPTNeoXModel.forward.attention_mask)
              for more details. Should be a tensor of shape `(batch_size, seq_len)` (same as `input_ids`) with
              0s for tokens that are masked and 1s for tokens that are not masked. This means it is given by
              `batch.code != batch.PAD_INDEX` as whenever the code is not a padding token, it should be
              attended to.

        Args:
            batch: The input batch of data.

        Returns:
            A dictionary of inputs for the Hugging Face model.

        Examples:
            >>> model = Model({
            ...     "num_hidden_layers": 2,
            ...     "num_attention_heads": 2,
            ...     "hidden_size": 4,
            ...     "max_position_embeddings": 3,
            ...     "vocab_size": 10,
            ... })
            >>> batch = Mock(code=torch.LongTensor([[0, 3, 1], [0, 0, 2]]), PAD_INDEX=0)
            >>> model._hf_inputs(batch)
            {'input_ids': tensor([[0, 3, 1],
                                  [0, 0, 2]]),
             'attention_mask': tensor([[False,  True,  True],
                                       [False, False,  True]])}
        """
        return {
            "input_ids": batch.code,
            "attention_mask": (batch.code != batch.PAD_INDEX),
        }

    def _forward_demo(self, batch: MEDSTorchBatch) -> tuple[torch.FloatTensor, CausalLMOutputWithPast]:
        """A demo forward pass that adds more checks and assertions."""

        self._check_inputs(batch)
        self._check_parameters()
        out = self._forward(batch)
        self._check_outputs(*out)

        return out

    def _forward(self, batch: MEDSTorchBatch) -> tuple[torch.FloatTensor, CausalLMOutputWithPast]:
        outputs = self.HF_model(**self._hf_inputs(batch))
        loss = F.cross_entropy(
            outputs.logits[:, :-1].transpose(2, 1), batch.code[:, 1:], ignore_index=batch.PAD_INDEX
        )

        return loss, outputs

    def generate(self, batch: MEDSTorchBatch, do_sample: bool = True, **kwargs) -> torch.Tensor:
        """Generates a sequence of tokens from the model using the HF generation mixin.

        Examples:
            >>> _ = torch.manual_seed(0)
            >>> torch.use_deterministic_algorithms(True)
            >>> model = Model({
            ...     "num_hidden_layers": 2,
            ...     "num_attention_heads": 2,
            ...     "hidden_size": 4,
            ...     "max_position_embeddings": 10,
            ...     "vocab_size": dataset_config.vocab_size,
            ... }, precision="32-true")

        This model has a maximum sequence length of 10. If we check, our sample batch has a sequence length of
        9:

            >>> sample_batch.code.shape
            torch.Size([2, 9])

        This means that by default, the model will generate 1 token:

            >>> print(model.generate(sample_batch, do_sample=False))
            tensor([[2],
                    [2]])

        If we create a model with a maximum sequence length of 20, we can generate 11 tokens:

            >>> _ = torch.manual_seed(0)
            >>> model = Model({
            ...     "num_hidden_layers": 2,
            ...     "num_attention_heads": 2,
            ...     "hidden_size": 4,
            ...     "max_position_embeddings": 20,
            ...     "vocab_size": dataset_config.vocab_size,
            ... }, precision="32-true")
            >>> print(model.generate(sample_batch, do_sample=False))
            tensor([[ 2, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16],
                    [ 2, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16]])

        Note that here, as we've turned sampling off for determinism in these testing algorithms, the
        generated samples are clearly visibly not meaningful. This is completely expected because we are
        working with a model that has just been randomly initialized. What if we try with a model that has
        undergone a (tiny) bit of pre-training? TODO(generation test).
        """

        for_hf = self._hf_inputs(batch)

        generation_config = GenerationConfig(
            max_new_tokens=self.max_seq_len - batch.code.shape[1],
            do_sample=do_sample,
            num_beams=1,  # no beam search
            temperature=1.0,
            pad_token_id=batch.PAD_INDEX,
            bos_token_id=None,
            eos_token_id=self.HF_model.config.eos_token_id,
        )

        output_ids = self.HF_model.generate(
            for_hf.pop("input_ids"),
            generation_config=generation_config,
            **for_hf,
            **kwargs,
        )

        input_seq_len = batch.code.shape[1]

        return output_ids[:, input_seq_len:]



import polars as pl
from meds_torchdata.config import MEDSTorchDataConfig


def get_timeline_end_token_idx(dataset_config: MEDSTorchDataConfig) -> int:
    """Get the index of the end token in the timeline vocabulary.

    Args:
        dataset (MEDSPytorchDataset): The dataset used for generation.

    Returns:
        int: The index of the end token in the timeline vocabulary.

    Examples:
        >>>
        37
    """
    columns = ["code", "code/vocab_index"]
    code_metadata_df = pl.read_parquet(dataset_config.code_metadata_fp, columns=columns, use_pyarrow=True)

    return code_metadata_df.filter(pl.col("code") == "TIMELINE//END").select("code/vocab_index").item()



import logging

import torch
from meds_torchdata import MEDSTorchBatch
from omegaconf import ListConfig
from torchmetrics import Metric, MetricCollection
from torchmetrics.text import Perplexity

logger = logging.getLogger(__name__)


class TopKMulticlassAccuracy(Metric):
    """Computes the fraction of samples that yield a correct prediction within their top-k class logits.

    The torchmetrics codebase Accuracy metric does not perform what I expect it to in top-k and multi-class
    settings. This metric behaves as expected.

    See https://github.com/Lightning-AI/torchmetrics/issues/3068 for tracking.

    Args:
        top_k: A prediction for a given sequence event is correct if the true label is among the top `top_k`
            predicted logits.
        ignore_index: The index to ignore when computing the accuracy. This is useful for ignoring padding
            indices. For this metric, it must be set.

    Returns:
        A (scalar) tensor with the accuracy of the top-k predictions. If there are no valid predictions (e.g.,
        all labels are padding), returns 0.0.

    Raises:
        ValueError: If the logits and target tensors do not have the same shape in the batch (0) and sequence
            (1) dimensions.

    Examples:
        >>> code = torch.LongTensor([[3, 2], [1, 0]])
        >>> logits = torch.FloatTensor(
        ...     [
        ...         [[0.0, 0.1, 0.5, 0.4], [0.0, 0.2, 0.7, 0.1]], # Prediction orders: 2, 3, 1; 2, 1, 3
        ...         [[0.0, 0.4, 0.3, 0.3], [1.0, 0.0, 0.0, 0.0]], # Prediction orders: 1, 2 & 3; padding
        ...     ]
        ... )
        >>> TopKMulticlassAccuracy(top_k=2, ignore_index=0)(logits, code)
        tensor(1.)
        >>> TopKMulticlassAccuracy(top_k=1, ignore_index=0)(logits, code)
        tensor(0.6667)
        >>> code = torch.LongTensor([[1, 1], [1, 0]])
        >>> TopKMulticlassAccuracy(top_k=1, ignore_index=0)(logits, code)
        tensor(0.3333)
        >>> TopKMulticlassAccuracy(top_k=2, ignore_index=0)(logits, code)
        tensor(0.6667)
        >>> TopKMulticlassAccuracy(top_k=3, ignore_index=0)(logits, code)
        tensor(1.)
        >>> code = torch.LongTensor([[1, 0], [0, 0]])
        >>> TopKMulticlassAccuracy(top_k=3, ignore_index=0)(logits, code)
        tensor(1.)
        >>> TopKMulticlassAccuracy(top_k=2, ignore_index=0)(logits, code)
        tensor(0.)

    If all labels are ignored, the accuracy is 0.0:

        >>> code = torch.LongTensor([[0, 0], [0, 0]])
        >>> TopKMulticlassAccuracy(top_k=2, ignore_index=0)(logits, code)
        tensor(0.)

    Errors are raised if the shapes are misaligned:

        >>> code = torch.LongTensor([[1, 1, 0], [0, 3, 0]])
        >>> TopKMulticlassAccuracy(top_k=2, ignore_index=0)(logits, code)
        Traceback (most recent call last):
            ...
        ValueError: logits and target must have the same shape in the batch (0) and sequence (1) dimensions.
            Got torch.Size([2, 2, 4]) and torch.Size([2, 3])
    """

    top_k: int
    ignore_index: int

    def __init__(self, top_k: int, ignore_index: int = 0, **kwargs):
        super().__init__(**kwargs)

        self.top_k = top_k
        self.ignore_index = ignore_index

        self.add_state("correct", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, logits: torch.FloatTensor, target: torch.LongTensor) -> None:
        batch, seq_len = target.shape
        if logits.shape[0] != batch or logits.shape[1] != seq_len:
            raise ValueError(
                "logits and target must have the same shape in the batch (0) and sequence (1) dimensions. "
                f"Got {logits.shape} and {target.shape}"
            )

        topk = torch.topk(logits, k=self.top_k, dim=2).indices  # batch x seq_len x top_k
        correct = (topk == target.unsqueeze(2)).any(dim=2)  # batch x seq_len

        keep_mask = target != self.ignore_index

        self.correct += correct[keep_mask].sum()
        self.total += keep_mask.sum()

    def compute(self) -> torch.FloatTensor:
        safe_denom = torch.where(self.total > 0, self.total, torch.ones_like(self.total))

        return torch.where(self.total > 0, self.correct.float() / safe_denom, torch.tensor(0.0))


class NextCodeMetrics(Metric):
    """A `torchmetrics` Metric for next code prediction in Autoregressive "Everything-is-code" models.

    This module is largely a simple wrapper around `torchmetrics.MetricCollection` to enable configuration and
    code isolation and to seamlessly slice the predictions and tokens to the right shapes.

    Supported metrics:
      - Top-$k$ accuracy
      - Perplexity

    Supported Vocabulary Subdivisions:
      - All codes

    Attributes:
        accuracies: The top-$k$ accuracy metrics contained in this metric collection.
        perplexity: The perplexity metric.

    Examples:
        >>> M = NextCodeMetrics(top_k=[1, 2, 3], vocab_size=4)

    To show it in use, we'll need some codes (targets) and logits (predictions):

        >>> code = torch.LongTensor([[1, 3, 2], [2, 1, 0]])
        >>> logits = torch.FloatTensor([
        ...    [
        ...        [0.0, 0.1, 0.5, 0.4], # Label is 3; Prediction order is 2, 3, 1
        ...        [0.0, 0.2, 0.7, 0.1], # Label is 2; Prediction order is 2, 1, 3
        ...        [0.0, 1.0, 0.0, 0.0], # No label; Should be dropped.
        ...    ], [
        ...        [0.0, 0.4, 0.3, 0.3], # Label is 1; Prediction order is 1, 2 & 3
        ...        [1.0, 0.0, 0.0, 0.0], # Padding label; Should be ignored.
        ...        [0.0, 0.0, 2.0, 0.0], # No label; Should be dropped.
        ...    ]
        ... ])

    We'll make a mock batch as this metric will just use the `code` attribute:

        >>> batch = Mock(spec=MEDSTorchBatch)
        >>> batch.code = code

    Then, we can update and compute the metric values:

        >>> M(logits, batch)
        {'Accuracy/top_1': tensor(0.6667), 'Accuracy/top_2': tensor(1.), 'Accuracy/top_3': tensor(1.),
         'perplexity': tensor(3.1896)}

    You can also run on a single `top_k`:

        >>> M = NextCodeMetrics(top_k=1, vocab_size=4)

    If `top_k` is not an int or a list of ints, an error is raised:

        >>> M = NextCodeMetrics(top_k=[1, "foo"], vocab_size=4)
        Traceback (most recent call last):
            ...
        ValueError: Invalid type for top_k. Want list[int] | int, got <class 'list'> ([1, 'foo']).

    If the max `top_k` is greater than the vocab size, a warning is logged and the `top_k` is filtered to only
    those valid `top_k` values:

        >>> with print_warnings():
        ...     M = NextCodeMetrics(top_k=[1, 2, 3], vocab_size=3)
        Warning: Top-k accuracy requested for k (3 >= vocab_size (3). This is not a valid metric. Filtering to
        only requested k < 3.
        >>> sorted(M.accuracies.keys())
        ['Accuracy/top_1', 'Accuracy/top_2']

    If no valid `top_k` is requested, a warning is logged and the `top_k` is set to 1:

        >>> with print_warnings():
        ...     M = NextCodeMetrics(top_k=[], vocab_size=2)
        Warning: No valid top-k accuracy requested. Adding top-k of 1.
        >>> sorted(M.accuracies.keys())
        ['Accuracy/top_1']
    """

    def __init__(self, top_k: list[int] | int, vocab_size: int, ignore_index: int = 0, **base_metric_kwargs):
        super().__init__(**base_metric_kwargs)

        match top_k:
            case int():
                top_k = [top_k]
            case list() | ListConfig() if all(isinstance(k, int) for k in top_k):
                pass
            case _:
                raise ValueError(
                    f"Invalid type for top_k. Want list[int] | int, got {type(top_k)} ({top_k})."
                )

        if top_k and (max(top_k) >= vocab_size):
            logger.warning(
                f"Top-k accuracy requested for k ({max(top_k)} >= vocab_size ({vocab_size}). "
                f"This is not a valid metric. Filtering to only requested k < {vocab_size}."
            )
            top_k = [k for k in top_k if k < vocab_size]

        if not top_k:
            logger.warning("No valid top-k accuracy requested. Adding top-k of 1.")
            top_k = [1]

        self.accuracies = MetricCollection(
            {f"Accuracy/top_{k}": TopKMulticlassAccuracy(top_k=k, ignore_index=ignore_index) for k in top_k}
        )
        self.perplexity = Perplexity(ignore_index=ignore_index)

        self.hparams = {
            "top_k": top_k,
            "vocab_size": vocab_size,
            "ignore_index": ignore_index,
            **base_metric_kwargs,
        }

    def update(self, logits: torch.Tensor, batch: MEDSTorchBatch):
        """Update the metric with the current batch and logits, sliced to match targets and predictions.

        Args:
            logits: The logits from the model, of shape (batch_size, sequence_length, vocab_size).
            batch: The MEDSTorchBatch containing the input codes at attribute `.code`.  Note that the `code`
                and logits are aligned such that the given code inputs are at the same position as the logits
                produced at that input -- so the logits need to be shifted to align with their prediction
                targets.
        """

        logits = logits[:, :-1]
        targets = batch.code[:, 1:]

        self.perplexity.update(logits, targets)
        self.accuracies.update(logits, targets)

    def compute(self):
        return {**self.accuracies.compute(), "perplexity": self.perplexity.compute()}



from pathlib import Path

import polars as pl
from meds import held_out_split, train_split, tuning_split
from polars.testing import assert_frame_equal


def test_generate_trajectories_runs(generated_trajectories: Path):
    generated_files = generated_trajectories.glob("*/*.parquet")
    assert len(list(generated_files)) > 0, "No generated files found in the specified directory."

    trajectories_by_split = {}
    for split in (train_split, tuning_split, held_out_split):
        split_dir = generated_trajectories / split

        trajectories_by_split[split] = {}

        for fp in split_dir.glob("*.parquet"):
            trajectories_by_split[split][fp.stem] = pl.read_parquet(fp, use_pyarrow=True)

    assert len(trajectories_by_split) == 3, "Not all splits have generated trajectories."

    for sp, samps in trajectories_by_split.items():
        assert len(samps) == 2, f"Expected 2 trajectories for split {sp}, but found {len(samps)}."

        try:
            assert_frame_equal(samps["0"], samps["1"], check_exact=True)
            samps_equal = True
        except AssertionError:
            samps_equal = False

        assert not samps_equal, f"Trajectories for distinct samples in split {sp} are equal!"

        subjects = {samp: set(df["subject_id"]) for samp, df in samps.items()}
        assert subjects["0"] == subjects["1"], f"Subjects in samples for split {sp} do not match!"



import copy
import logging
import os
import subprocess
from importlib.resources import files
from pathlib import Path

import hydra
from omegaconf import DictConfig

logger = logging.getLogger(__name__)

CONFIGS = files("MEDS_EIC_AR") / "preprocessing" / "configs"


@hydra.main(version_base=None, config_path=str(CONFIGS), config_name="_process_data")
def process_data(cfg: DictConfig):
    input_dir = Path(cfg.input_dir)
    intermediate_dir = Path(cfg.intermediate_dir)
    output_dir = Path(cfg.output_dir)
    do_demo = cfg.do_demo

    # 0. Pre-MTD pre-processing
    logger.info("Pre-MTD pre-processing")
    done_fp = intermediate_dir / ".done"
    if done_fp.exists():  # pragma: no cover
        logger.info("Pre-MTD pre-processing already done, skipping")
    else:
        env = copy.deepcopy(os.environ)
        env["RAW_MEDS_DIR"] = str(input_dir)
        env["MTD_INPUT_DIR"] = str(intermediate_dir)

        if do_demo:
            env["MIN_SUBJECTS_PER_CODE"] = "2"
            env["MIN_EVENTS_PER_SUBJECT"] = "1"

        pipeline_config_fp = (CONFIGS / "_reshard_data.yaml") if cfg.do_reshard else (CONFIGS / "_data.yaml")
        cmd = [
            "MEDS_transform-pipeline",
            f"pipeline_config_fp={pipeline_config_fp!s}",
        ]
        logger.info(f"Running command: {' '.join(cmd)}")

        result = subprocess.run(cmd, env=env, capture_output=True, check=False)
        if result.returncode != 0:  # pragma: no cover
            logger.error("Error running MEDS_transform-pipeline")
            logger.error(result.stdout.decode())
            logger.error(result.stderr.decode())
            raise RuntimeError("Error running MEDS_transform-pipeline")

        logger.info("Pre-MTD pre-processing done")
        done_fp.touch()

    # 1. Run MTD pre-processing
    logger.info("Running MTD pre-processing")
    done_fp = output_dir / ".done"

    if done_fp.exists():  # pragma: no cover
        logger.info("MTD pre-processing already done, skipping")
    else:
        env = copy.deepcopy(os.environ)

        cmd = [
            "MTD_preprocess",
            f"MEDS_dataset_dir={intermediate_dir!s}",
            f"output_dir={output_dir!s}",
        ]
        logger.info(f"Running command: {' '.join(cmd)}")

        result = subprocess.run(cmd, env=env, capture_output=True, check=False)
        if result.returncode != 0:  # pragma: no cover
            logger.error("Error running MTD_preprocess")
            logger.error(result.stderr.decode())
            raise RuntimeError("Error running MTD_preprocess")

        logger.info("MTD pre-processing done")
        done_fp.touch()

    return



from pathlib import Path

from meds_torchdata import MEDSPytorchDataset


def test_process_data_runs(preprocessed_dataset: Path):
    out_files = list(preprocessed_dataset.rglob("*.*"))
    assert len(out_files) > 0


def test_process_dataset_correct(pytorch_dataset: MEDSPytorchDataset):
    assert len(pytorch_dataset) > 0


def test_process_with_reshard_correct(preprocessed_dataset_with_reshard: Path):
    out_files = list(preprocessed_dataset_with_reshard.rglob("*.*"))
    assert len(out_files) > 0



"""Test set-up and fixtures code."""

import subprocess
import tempfile
from contextlib import contextmanager
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock, Mock, patch

import pytest
from meds_testing_helpers.dataset import MEDSDataset
from meds_torchdata import MEDSPytorchDataset, MEDSTorchBatch, MEDSTorchDataConfig
from torch.utils.data import DataLoader

from MEDS_EIC_AR.model.model import Model
from MEDS_EIC_AR.training.module import MEICARModule


def run_and_check(cmd: list[str]) -> dict[str, str]:
    """Runs a command and checks the output.

    Args:
        cmd: The command to run. This should be a list of strings, which will be run through subprocess.run in
            shell=False mode.

    Raises:
        ValueError: If the command fails.

    Returns:
        A dictionary with the keys "stdout" and "stderr", containing the output of the command.
    """

    cmd_out = subprocess.run(cmd, capture_output=True, check=False)

    out = {"stdout": cmd_out.stdout.decode(), "stderr": cmd_out.stderr.decode()}

    if cmd_out.returncode == 0:
        return

    err = [f"Command yielded code {cmd_out.returncode}", "Stdout:", out["stdout"], "Stderr:", out["stderr"]]
    raise ValueError("\n".join(err))


def run_process_data(input_dir: Path, root_dir: Path, do_demo: bool = True, do_reshard: bool = False) -> Path:
    """Runs the pre-process code and returns the output directory.

    Args:
        input_dir: The input directory to process.
        root_dir: This is the root dir of the output data, and the intermediate dir is set to
            `root_dir / "intermediate"` and the output dir is set to `root_dir / "output"`.
        do_demo: Whether to run the demo mode.
        do_reshard: Whether to run the resharding step.

    Returns:
        The output directory (`root_dir / "output"`).

    Raises:
        ValueError: If the command fails.
    """

    interemediate_dir = root_dir / "intermediate"
    output_dir = root_dir / "output"

    run_and_check(
        [
            "MEICAR_process_data",
            f"input_dir={input_dir!s}",
            f"intermediate_dir={interemediate_dir!s}",
            f"output_dir={output_dir!s}",
            f"do_demo={do_demo}",
            f"do_reshard={do_reshard}",
        ]
    )

    return output_dir


@pytest.fixture(scope="session")
def preprocessed_dataset_with_reshard(
    simple_static_MEDS: Path, tmp_path_factory: pytest.TempPathFactory
) -> Path:
    """Fixture to create a preprocessed dataset."""

    test_root = tmp_path_factory.mktemp("preprocessed_dataset_with_reshard")

    return run_process_data(simple_static_MEDS, test_root, do_reshard=True)


@pytest.fixture(scope="session")
def preprocessed_dataset(simple_static_MEDS: Path, tmp_path_factory: pytest.TempPathFactory) -> Path:
    """Fixture to create a preprocessed dataset."""

    test_root = tmp_path_factory.mktemp("preprocessed_dataset")

    return run_process_data(simple_static_MEDS, test_root)


@pytest.fixture(scope="session")
def preprocessed_dataset_with_task(
    preprocessed_dataset: Path,
    simple_static_MEDS_dataset_with_task: Path,
) -> tuple[Path, Path, str]:
    D = MEDSDataset(root_dir=simple_static_MEDS_dataset_with_task)

    if len(D.task_names) != 1:  # pragma: no cover
        raise ValueError("Expected only one task in the dataset.")

    yield preprocessed_dataset, D.task_root_dir, D.task_names[0]


@pytest.fixture(scope="session")
def dataset_config(preprocessed_dataset: Path) -> MEDSTorchDataConfig:
    """Fixture to create a dataset configuration."""
    return MEDSTorchDataConfig(tensorized_cohort_dir=preprocessed_dataset, max_seq_len=10)


@pytest.fixture(scope="session")
def dataset_config_with_task(preprocessed_dataset_with_task: tuple[Path, Path, str]) -> MEDSTorchDataConfig:
    """Fixture to create a dataset configuration."""
    cohort_dir, tasks_dir, task_name = preprocessed_dataset_with_task
    return MEDSTorchDataConfig(
        tensorized_cohort_dir=cohort_dir,
        max_seq_len=10,
        task_labels_dir=(tasks_dir / task_name),
        seq_sampling_strategy="to_end",
    )


@pytest.fixture(scope="session")
def pytorch_dataset(dataset_config: MEDSTorchDataConfig) -> MEDSPytorchDataset:
    """Fixture to create a PyTorch dataset."""
    return MEDSPytorchDataset(dataset_config, split="train")


@pytest.fixture(scope="session")
def pytorch_dataset_with_task(dataset_config_with_task: MEDSTorchDataConfig) -> MEDSPytorchDataset:
    """Fixture to create a PyTorch dataset with task labels."""
    return MEDSPytorchDataset(dataset_config_with_task, split="train")


@pytest.fixture(scope="session")
def sample_batch(pytorch_dataset: MEDSPytorchDataset) -> MEDSTorchBatch:
    """Fixture to create a sample batch."""
    dataloader = DataLoader(pytorch_dataset, batch_size=2, shuffle=False, collate_fn=pytorch_dataset.collate)
    return list(dataloader)[1]


@pytest.fixture(scope="session")
def pretrained_model(preprocessed_dataset: Path, tmp_path_factory: pytest.TempPathFactory) -> Path:
    output_dir = tmp_path_factory.mktemp("pretrained_model")

    run_and_check(
        [
            "MEICAR_pretrain",
            "--config-name=_demo_pretrain",
            f"output_dir={output_dir!s}",
            f"datamodule.config.tensorized_cohort_dir={preprocessed_dataset!s}",
        ]
    )

    return output_dir


@pytest.fixture(scope="session")
def pretrained_module(pretrained_model: Path) -> MEICARModule:
    """Returns the pre-trained MEICAR Lightning Module."""

    ckpt_path = pretrained_model / "best_model.ckpt"
    if not ckpt_path.is_file():
        raise ValueError("No best checkpoint reported.")

    return MEICARModule.load_from_checkpoint(ckpt_path)


@pytest.fixture(scope="session")
def pretrained_GPT_model(pretrained_module: MEICARModule) -> Model:
    """Returns the HF model backbone of the pre-trained MEICAR Lightning Module."""

    return pretrained_module.model


@pytest.fixture(scope="session")
def generated_trajectories(
    pretrained_model: Path,
    preprocessed_dataset_with_task: tuple[Path, Path, str],
    tmp_path_factory: pytest.TempPathFactory,
) -> Path:
    tensorized_cohort_dir, task_root_dir, task_name = preprocessed_dataset_with_task
    model_initialization_dir = pretrained_model

    output_dir = tmp_path_factory.mktemp("generated_trajectories")

    run_and_check(
        [
            "MEICAR_generate_trajectories",
            "--config-name=_demo_generate_trajectories",
            f"output_dir={output_dir!s}",
            f"model_initialization_dir={model_initialization_dir!s}",
            f"datamodule.config.tensorized_cohort_dir={tensorized_cohort_dir!s}",
            f"datamodule.config.task_labels_dir={(task_root_dir / task_name)!s}",
            "datamodule.batch_size=2",
            "trainer=demo",
        ]
    )

    return output_dir


@contextmanager
def print_warnings(caplog: pytest.LogCaptureFixture):
    """Captures all logged warnings within this context block and prints them upon exit.

    This is useful in doctests, where you want to show printed outputs for documentation and testing purposes.
    """

    n_current_records = len(caplog.records)

    with caplog.at_level("WARNING"):
        yield
    # Print all captured warnings upon exit
    for record in caplog.records[n_current_records:]:
        print(f"Warning: {record.getMessage()}")


@pytest.fixture(autouse=True)
def _setup_doctest_namespace(
    doctest_namespace: dict[str, Any],
    caplog: pytest.LogCaptureFixture,
    simple_static_MEDS: Path,
    simple_static_MEDS_dataset_with_task: Path,
    sample_batch: MEDSTorchBatch,
    preprocessed_dataset: Path,
    dataset_config: MEDSTorchDataConfig,
    pretrained_GPT_model: Model,
    pretrained_module: MEICARModule,
    pytorch_dataset: MEDSPytorchDataset,
    pytorch_dataset_with_task: MEDSPytorchDataset,
):
    doctest_namespace.update(
        {
            "print_warnings": partial(print_warnings, caplog),
            "patch": patch,
            "MagicMock": MagicMock,
            "Path": Path,
            "Mock": Mock,
            "datetime": datetime,
            "tempfile": tempfile,
            "simple_static_MEDS": simple_static_MEDS,
            "simple_static_MEDS_dataset_with_task": simple_static_MEDS_dataset_with_task,
            "preprocessed_dataset": preprocessed_dataset,
            "sample_batch": sample_batch,
            "dataset_config": dataset_config,
            "pretrained_GPT_model": pretrained_GPT_model,
            "pretrained_module": pretrained_module,
            "pytorch_dataset": pytorch_dataset,
            "pytorch_dataset_with_task": pytorch_dataset_with_task,
        }
    )



from hashlib import sha256

from MEDS_transforms.configs.utils import OmegaConfResolver
from omegaconf import DictConfig


def hash_based_seed(seed: int | None, split: str, sample: int) -> int:
    """Generates a hash-based seed for reproducibility.

    This function generates a hash-based seed using the provided seed, split, and sample values. It is
    designed to be used in conjunction with OmegaConf for configuration management.

    Args:
        seed: The original seed value. THIS WILL NOT OVERWRITE THE OUTPUT. Rather, this just ensures the
            sequence of seeds chosen can be deterministically updated by changing a base parameter.
        split: The split identifier.
        sample: The sample index.

    Returns:
        A hash-based seed value.

    Examples:
        >>> hash_based_seed(42, "train", 0)
        1508872876
        >>> hash_based_seed(None, "held_out", 1)
        3132876237
    """

    hash_str = f"{seed}_{split}_{sample}"
    return int(sha256(hash_str.encode()).hexdigest(), 16) % (2**32 - 1)


@OmegaConfResolver
def int_prod(x: int, y: int) -> int:
    """Returns the closest integer to the product of x and y.

    This function can be used in omega conf configs as a resolved function.

    Args:
        x: The first integer.
        y: The second integer.

    Returns:
        The closest integer to the product of x and y.

    Examples:
        >>> int_prod(2, 3)
        6
        >>> int_prod(2, 3.5)
        7
        >>> int_prod(2.49, 3)
        7
    """
    return round(x * y)


@OmegaConfResolver
def resolve_generation_context_size(seq_lens: DictConfig) -> int:
    """Resolves the target generation context (input) size for the model.

    This function can be used in omega conf configs as a resolved function.

    Args:
        seq_lens: A configuration object containing the following key/value pairs:
            - max_generated_trajectory_len: If set, this gives the maximum length of trajectories (outputs)
              that should be generated.
            - frac_seq_len_as_context: If set, this gives the fraction of the pre-trained model's maximum
              sequence length that should be used as the context (input) for generation.
            - generation_context_size: If set, this gives the exact context size to use for generation.
            - pretrained_max_seq_len: The maximum sequence length of the pre-trained model.

    Returns:
        The generation context size, which is the maximum length of the input sequences the dataloader will
        pass to the model. The remaining length of the sequence will be used for generation. This will take
        one of several values depending on what is set:
            - If `generation_context_size` is set, it is returned.
            - If `max_generated_trajectory_len` is set, then
              `pretrained_max_seq_len - max_generated_trajectory_len` is returned.
            - If `frac_seq_len_as_context` is set, then
              `round(pretrained_max_seq_len * frac_seq_len_as_context)` is returned.

    Raises:
        TypeError: If the input keys have the wrong types.
        ValueError: If none of `max_generated_trajectory_len`, `frac_seq_len_as_context`, or
            `generation_context_size` are set, if more than one of them are set, if
            `pretrained_max_seq_len` is not set, or if the returned value would not be a positive integer.

    Examples:
        >>> resolve_generation_context_size(
        ...     {"pretrained_max_seq_len": 1024, "max_generated_trajectory_len": 512}
        ... )
        512
        >>> resolve_generation_context_size(
        ...     {"pretrained_max_seq_len": 1024, "generation_context_size": 100}
        ... )
        100
        >>> resolve_generation_context_size(
        ...     {"pretrained_max_seq_len": 1024, "frac_seq_len_as_context": 0.75}
        ... )
        768

    Fractional resolution is guaranteed to never be greater than the maximum sequence length or less than 1:

        >>> resolve_generation_context_size(
        ...     {"pretrained_max_seq_len": 1024, "frac_seq_len_as_context": 0.9999999999}
        ... )
        1023
        >>> resolve_generation_context_size(
        ...     {"pretrained_max_seq_len": 1024, "frac_seq_len_as_context": 0.0000000001}
        ... )
        1

    Null values do not trigger errors nor are used:

        >>> resolve_generation_context_size(
        ...     {
        ...         "pretrained_max_seq_len": 1024, "frac_seq_len_as_context": 0.75,
        ...         "generation_context_size": None
        ...     }
        ... )
        768

    Errors are raised if the input is missing required keys...

        >>> resolve_generation_context_size({})
        Traceback (most recent call last):
            ...
        ValueError: Required key 'pretrained_max_seq_len' not found in input.
        >>> resolve_generation_context_size({"pretrained_max_seq_len": 1024})
        Traceback (most recent call last):
            ...
        ValueError: Exactly one of 'max_generated_trajectory_len' or 'frac_seq_len_as_context' or
            'generation_context_size' must be set to a non-null value.

    or if it has too many keys...

        >>> resolve_generation_context_size(
        ...     {
        ...         "pretrained_max_seq_len": 1024, "frac_seq_len_as_context": 0.75,
        ...         "generation_context_size": 256
        ...     }
        ... )
        Traceback (most recent call last):
            ...
        ValueError: Exactly one of 'max_generated_trajectory_len' or 'frac_seq_len_as_context' or
            'generation_context_size' must be set to a non-null value.

    or if it has extra keys...
        >>> resolve_generation_context_size(
        ...     {
        ...         "pretrained_max_seq_len": 1024, "frac_seq_len_as_context": 0.75,
        ...         "foobar": 256
        ...     }
        ... )
        Traceback (most recent call last):
            ...
        ValueError: Extra keys found in input: ['foobar']. Only 'max_generated_trajectory_len',
            'frac_seq_len_as_context', 'generation_context_size', 'pretrained_max_seq_len' are allowed.

    or if the keys have the wrong types:

        >>> resolve_generation_context_size(
        ...     {"pretrained_max_seq_len": 1024, "generation_context_size": "foobar"}
        ... )
        Traceback (most recent call last):
            ...
        TypeError: Expected 'generation_context_size' to be an int; got <class 'str'>.
        >>> resolve_generation_context_size(
        ...     {"pretrained_max_seq_len": 1024, "max_generated_trajectory_len": -10}
        ... )
        Traceback (most recent call last):
            ...
        ValueError: Expected 'max_generated_trajectory_len' to be positive; got -10.
        >>> resolve_generation_context_size(
        ...     {"pretrained_max_seq_len": 1024, "frac_seq_len_as_context": 1.25}
        ... )
        Traceback (most recent call last):
            ...
        ValueError: If non-null, 'frac_seq_len_as_context' must be a float between 0 and 1. Got 1.25.

    Errors are also raised if the output would not be a positive integer:

        >>> resolve_generation_context_size({"pretrained_max_seq_len": 1, "max_generated_trajectory_len": 5})
        Traceback (most recent call last):
            ...
        ValueError: The maximum sequence length of the pre-trained model must be greater than the maximum
            generated trajectory length. Got 1 and 5.
    """

    if seq_lens.get("pretrained_max_seq_len", None) is None:
        raise ValueError("Required key 'pretrained_max_seq_len' not found in input.")

    allowed_keys = [
        "max_generated_trajectory_len",
        "frac_seq_len_as_context",
        "generation_context_size",
        "pretrained_max_seq_len",
    ]

    if extra_keys := set(seq_lens.keys()) - set(allowed_keys):
        allowed_keys_str = "', '".join(allowed_keys)
        raise ValueError(
            f"Extra keys found in input: {sorted(extra_keys)}. Only '{allowed_keys_str}' are allowed."
        )

    non_null_keys = {k: v for k, v in seq_lens.items() if v is not None}
    pretrained_seq_len = non_null_keys.pop("pretrained_max_seq_len")

    if len(non_null_keys) != 1:
        raise ValueError(
            "Exactly one of 'max_generated_trajectory_len' or 'frac_seq_len_as_context' or "
            "'generation_context_size' must be set to a non-null value."
        )

    for k in ["pretrained_max_seq_len", "max_generated_trajectory_len", "generation_context_size"]:
        if k not in non_null_keys:
            continue
        if not isinstance(seq_lens[k], int):
            raise TypeError(f"Expected '{k}' to be an int; got {type(seq_lens[k])}.")
        if seq_lens[k] <= 0:
            raise ValueError(f"Expected '{k}' to be positive; got {seq_lens[k]}.")

    if "generation_context_size" in non_null_keys:
        return non_null_keys["generation_context_size"]
    if "max_generated_trajectory_len" in non_null_keys:
        if pretrained_seq_len <= non_null_keys["max_generated_trajectory_len"]:
            raise ValueError(
                "The maximum sequence length of the pre-trained model must be greater than the maximum "
                f"generated trajectory length. Got {pretrained_seq_len} and "
                f"{non_null_keys['max_generated_trajectory_len']}."
            )
        return pretrained_seq_len - non_null_keys["max_generated_trajectory_len"]
    if "frac_seq_len_as_context" in non_null_keys:
        val = non_null_keys["frac_seq_len_as_context"]
        if not isinstance(val, float) or val < 0 or val > 1:
            raise ValueError(
                f"If non-null, 'frac_seq_len_as_context' must be a float between 0 and 1. Got {val}."
            )
        return min(max(round(pretrained_seq_len * val), 1), pretrained_seq_len - 1)



import copy
import logging
import re
from collections.abc import Callable, Iterator
from functools import partial
from pathlib import Path
from typing import Any, Literal

import hydra
import lightning as L
import torch
from meds import held_out_split, train_split, tuning_split
from meds_torchdata import MEDSTorchBatch
from transformers.modeling_outputs import CausalLMOutputWithPast

from ..model import Model
from .metrics import NextCodeMetrics

logger = logging.getLogger(__name__)


def _factory_to_dict(factory: partial | None) -> dict[str, Any] | None:
    """Extracts a sufficient dictionary for reconstructing the optimizer or LR scheduler.

    Args:
        factory: A partial function that creates an optimizer or LR scheduler. This comes from the Hydra
            instantiation's "partial" functionality, which is used to partially initialize optimizers or LR
            schedulers without the need to pass in the parameters or optimizers, respectively.

    Returns:
        A dictionary suitable for storing, logging, and sufficient to reconstruct the given factory function.
        The dictionary will contain the special key "_target_" which contains the full path to the target
        function or module to be called. The rest of the dictionary will contain the keyword arguments passed
        in the partial. If the factory is None, returns None.

    Raises:
        TypeError: If the factory is not a partial function.
        ValueError: If the factory partial has any positional arguments or if it uses the reserved key
            "_target_" as a keyword argument.

    Examples:
        >>> print(_factory_to_dict(None))
        None
        >>> _factory_to_dict(partial(torch.optim.Adam, lr=0.001))
        {'_target_': 'torch.optim.adam.Adam', 'lr': 0.001}
        >>> from transformers import get_cosine_schedule_with_warmup
        >>> _factory_to_dict(partial(get_cosine_schedule_with_warmup, num_warmup_steps=10))
        {'_target_': 'transformers.optimization.get_cosine_schedule_with_warmup', 'num_warmup_steps': 10}

    Errors include type checking and some value checks:

        >>> _factory_to_dict(43)
        Traceback (most recent call last):
            ...
        TypeError: Expected a partial function, got <class 'int'>
        >>> _factory_to_dict(partial(torch.optim.Adam, 0.001))
        Traceback (most recent call last):
            ...
        ValueError: Expected a partial function with no positional arguments. Got (0.001,)
        >>> _factory_to_dict(partial(torch.optim.Adam, lr=0.001, _target_="foo"))
        Traceback (most recent call last):
            ...
        ValueError: Expected a partial function with no _target_ keyword argument. Got _target_=foo
    """
    if factory is None:
        return None

    if not isinstance(factory, partial):
        raise TypeError(f"Expected a partial function, got {type(factory)}")

    if factory.args:
        raise ValueError(f"Expected a partial function with no positional arguments. Got {factory.args}")

    kwargs = factory.keywords.copy()

    if "_target_" in kwargs:
        raise ValueError(
            "Expected a partial function with no _target_ keyword argument. "
            f"Got _target_={kwargs['_target_']}"
        )

    target = f"{factory.func.__module__}.{factory.func.__qualname__}"

    return {"_target_": target, **kwargs}


def _dict_to_factory(d: dict[str, Any] | None) -> partial:
    """Reconstructs a partial function from a dictionary.

    This is actually just a wrapper around `hydra.utils.instantiate` that sets the `_partial_` flag to True,
    so that it is clear we can use `_factory_to_dict` and `_dict_to_factory` to round-trip encode-decode the
    partial objects instantiated by `hydra.utils.instantiate`.

    Args:
        d: A dictionary containing the target function or module to be called under the key "_target_". The
            rest of the dictionary should contain the keyword arguments to be passed to the function.

    Returns:
        A partial function that creates an optimizer or LR scheduler.

    Examples:
        >>> d = {'_target_': 'torch.optim.adam.Adam', 'lr': 0.001}
        >>> factory = _dict_to_factory(d)
        >>> print(factory.func)
        <class 'torch.optim.adam.Adam'>
        >>> print(factory.keywords)
        {'lr': 0.001}
        >>> print(_factory_to_dict(factory))
        {'_target_': 'torch.optim.adam.Adam', 'lr': 0.001}
        >>> print(_dict_to_factory(None))
        None
    """

    return None if d is None else hydra.utils.instantiate(d, _partial_=True)


class MEICARModule(L.LightningModule):
    """A LightningModule for training and evaluating the MEICAR model."""

    @classmethod
    def load_from_checkpoint(cls, ckpt_path: Path | None = None) -> "MEICARModule":
        """Loads the full lightning module from a checkpoint.

        Args:
            ckpt_path: Path to the checkpoint file.

        Returns:
            The loaded MEICARModule instance, with all hyperparameters matching _except_ for the optimizer
            factory and LR scheduler which are discarded, as we can't tell from the saved data alone what
            classes they should be.

        Raises:
            KeyError: If the checkpoint does not contain the expected hyperparameters.

        Examples:
            >>> model = Model({
            ...     "num_hidden_layers": 2,
            ...     "num_attention_heads": 2,
            ...     "hidden_size": 4,
            ...     "max_position_embeddings": 3,
            ...     "vocab_size": 10,
            ... })
            >>> metrics = NextCodeMetrics(top_k=[1, 2, 3], vocab_size=4)
            >>> module = MEICARModule(model=model, metrics=metrics, optimizer=None)

        In pytorch lightning, saving and loading checkpoints is done using the `Trainer` class. We'll make one
        and attach the module to it for testing purposes:

            >>> trainer = L.Trainer(logger=False)
            >>> trainer.strategy.connect(module)

        We'll grab the models current parameters so we can compare after loading

            >>> import copy
            >>> model_params = copy.deepcopy(module.state_dict())

        Now, we can save the checkpoint to a temporary file and load it back in:

            >>> with tempfile.NamedTemporaryFile(suffix=".ckpt") as f:
            ...     trainer.save_checkpoint(f.name)
            ...     loaded_module = MEICARModule.load_from_checkpoint(f.name)

        We can check that the loaded module has the same parameters as the original:

            >>> if loaded_module.state_dict().keys() != model_params.keys():
            ...     print("Loaded module has different parameter names than the original!")
            ... else:
            ...     print("Loaded module has the same parameter names as the original!")
            Loaded module has the same parameter names as the original!
            >>> for k, v in model_params.items():
            ...     assert torch.equal(v, loaded_module.state_dict()[k]), f"Parameter {k} does not match"
        """

        checkpoint = torch.load(ckpt_path, map_location="cpu", weights_only=False)

        hparams = checkpoint.get("hyper_parameters", {})

        for k in ["model", "metrics", "optimizer", "LR_scheduler"]:
            if k not in hparams:
                raise KeyError(f"Checkpoint does not contain {k} hyperparameters. Got {list(hparams.keys())}")

        model = Model(**hparams["model"])
        metrics = NextCodeMetrics(**hparams["metrics"])
        optimizer = _dict_to_factory(hparams["optimizer"])
        LR_scheduler = _dict_to_factory(hparams["LR_scheduler"])

        return super().load_from_checkpoint(
            ckpt_path, model=model, metrics=metrics, optimizer=optimizer, LR_scheduler=LR_scheduler
        )

    def __init__(
        self,
        model: Model,
        metrics: NextCodeMetrics,
        optimizer: Callable[[Iterator[torch.nn.parameter.Parameter]], torch.optim.Optimizer] | None = None,
        LR_scheduler: Callable[[torch.optim.Optimizer], torch.optim.lr_scheduler._LRScheduler] | None = None,
    ):
        super().__init__()
        self.model = model
        self.metrics = metrics
        self.optimizer_factory = optimizer
        self.LR_scheduler_factory = LR_scheduler

        self.save_hyperparameters(
            {
                "model": model.hparams,
                "metrics": metrics.hparams,
                "optimizer": _factory_to_dict(self.optimizer_factory),
                "LR_scheduler": _factory_to_dict(self.LR_scheduler_factory),
            }
        )

    def _log_metrics(
        self,
        loss: torch.Tensor,
        outputs: CausalLMOutputWithPast,
        batch: MEDSTorchBatch,
        stage: Literal[train_split, tuning_split, held_out_split],
    ):
        batch_size = batch.batch_size

        is_train = stage == train_split

        self.log(f"{stage}/loss", loss, on_step=is_train, on_epoch=True, prog_bar=True, batch_size=batch_size)
        self.log_dict(
            {f"{stage}/{k}": v for k, v in self.metrics(outputs.logits, batch).items()},
            batch_size=batch_size,
            on_step=is_train,
        )

    def training_step(self, batch: MEDSTorchBatch):
        loss, outputs = self.model(batch)
        self._log_metrics(loss, outputs, batch, train_split)

        return loss

    def validation_step(self, batch):
        loss, outputs = self.model(batch)
        self._log_metrics(loss, outputs, batch, tuning_split)

        return loss

    @staticmethod
    def _is_norm_bias_param(n: str) -> bool:
        """Checks if a parameter name corresponds to a bias or normalization layer.

        Args:
            n: The name of the parameter.

        Returns:
            True if the parameter is a bias or normalization layer, False otherwise.

        Examples:
            >>> MEICARModule._is_norm_bias_param("model.decoder.bias")
            True
            >>> MEICARModule._is_norm_bias_param("model.layernorm.weight")
            True
            >>> MEICARModule._is_norm_bias_param("model.LayerNorm12.weight")
            True
            >>> MEICARModule._is_norm_bias_param("model.decoder.weight")
            False
            >>> MEICARModule._is_norm_bias_param("model.HF_model.gpt_neox.final_layer_norm.weight")
            True
        """
        return bool(re.search(r"(bias|layer(_?)norm(\d*)\.weight)", n, re.IGNORECASE))

    def _norm_bias_param_names(self) -> Iterator[str]:
        """Yields the names of parameters corresponding to the bias and normalization layers.

        These parameters should not be subject to weight decay by the optimizer.

        Examples:
            >>> list(pretrained_module._norm_bias_param_names())
            ['model.HF_model.gpt_neox.layers.0.input_layernorm.weight',
             'model.HF_model.gpt_neox.layers.0.input_layernorm.bias',
             'model.HF_model.gpt_neox.layers.0.post_attention_layernorm.weight',
             'model.HF_model.gpt_neox.layers.0.post_attention_layernorm.bias',
             'model.HF_model.gpt_neox.layers.0.attention.query_key_value.bias',
             'model.HF_model.gpt_neox.layers.0.attention.dense.bias',
             'model.HF_model.gpt_neox.layers.0.mlp.dense_h_to_4h.bias',
             'model.HF_model.gpt_neox.layers.0.mlp.dense_4h_to_h.bias',
             'model.HF_model.gpt_neox.layers.1.input_layernorm.weight',
             'model.HF_model.gpt_neox.layers.1.input_layernorm.bias',
             'model.HF_model.gpt_neox.layers.1.post_attention_layernorm.weight',
             'model.HF_model.gpt_neox.layers.1.post_attention_layernorm.bias',
             'model.HF_model.gpt_neox.layers.1.attention.query_key_value.bias',
             'model.HF_model.gpt_neox.layers.1.attention.dense.bias',
             'model.HF_model.gpt_neox.layers.1.mlp.dense_h_to_4h.bias',
             'model.HF_model.gpt_neox.layers.1.mlp.dense_4h_to_h.bias',
             'model.HF_model.gpt_neox.final_layer_norm.weight',
             'model.HF_model.gpt_neox.final_layer_norm.bias']
        """

        for name, _ in self.named_parameters():
            if self._is_norm_bias_param(name):
                yield name

    def _norm_bias_params(self) -> Iterator[torch.nn.parameter.Parameter]:
        """Yields the parameters corresponding to the bias and normalization layers."""

        for name in self._norm_bias_param_names():
            yield self.get_parameter(name)

    def _non_norm_bias_param_names(self) -> Iterator[str]:
        """Yields the names of parameters corresponding to the non-bias and non-normalization layers.

        These parameters should be subject to weight decay by the optimizer.

        Examples:
            >>> list(pretrained_module._non_norm_bias_param_names())
            ['model.HF_model.gpt_neox.embed_in.weight',
             'model.HF_model.gpt_neox.layers.0.attention.query_key_value.weight',
             'model.HF_model.gpt_neox.layers.0.attention.dense.weight',
             'model.HF_model.gpt_neox.layers.0.mlp.dense_h_to_4h.weight',
             'model.HF_model.gpt_neox.layers.0.mlp.dense_4h_to_h.weight',
             'model.HF_model.gpt_neox.layers.1.attention.query_key_value.weight',
             'model.HF_model.gpt_neox.layers.1.attention.dense.weight',
             'model.HF_model.gpt_neox.layers.1.mlp.dense_h_to_4h.weight',
             'model.HF_model.gpt_neox.layers.1.mlp.dense_4h_to_h.weight',
             'model.HF_model.embed_out.weight']
        """

        for name, _ in self.named_parameters():
            if not self._is_norm_bias_param(name):
                yield name

    def _non_norm_bias_params(self) -> Iterator[torch.nn.parameter.Parameter]:
        """Yields the parameters corresponding to the non-bias and non-normalization layers."""

        for name in self._non_norm_bias_param_names():
            yield self.get_parameter(name)

    @property
    def weight_decay(self) -> float | None:
        """Returns the weight decay value for the optimizer.

        This is used to set the weight decay value for the optimizer. If the optimizer factory does not
        contain a weight decay parameter, this will return None.

        Examples:
            >>> opt_factory = _dict_to_factory({"_target_": "torch.optim.adam.Adam", "weight_decay": 0.01})
            >>> model = Model({
            ...     "num_hidden_layers": 2,
            ...     "num_attention_heads": 2,
            ...     "hidden_size": 4,
            ...     "max_position_embeddings": 3,
            ...     "vocab_size": 10,
            ... })
            >>> metrics = NextCodeMetrics(top_k=[1, 2, 3], vocab_size=4)
            >>> MEICARModule(model=model, metrics=metrics, optimizer=opt_factory).weight_decay
            0.01
        """
        if self.optimizer_factory is None:
            return None
        else:
            return _factory_to_dict(self.optimizer_factory).get("weight_decay", None)

    @property
    def optimizer_no_decay_factory(
        self,
    ) -> Callable[[Iterator[torch.nn.parameter.Parameter]], torch.optim.Optimizer]:
        """Returns a factory function for creating an optimizer with no weight decay.

        This function is used to create an optimizer that does not apply weight decay to the bias and
        normalization parameters of the model. It is identical to the main optimizer factory, but with weight
        decay set to 0.0

        Examples:
            >>> opt_factory = _dict_to_factory({"_target_": "torch.optim.adam.Adam", "weight_decay": 0.01})
            >>> model = Model({
            ...     "num_hidden_layers": 2,
            ...     "num_attention_heads": 2,
            ...     "hidden_size": 4,
            ...     "max_position_embeddings": 3,
            ...     "vocab_size": 10,
            ... })
            >>> metrics = NextCodeMetrics(top_k=[1, 2, 3], vocab_size=4)
            >>> module = MEICARModule(model=model, metrics=metrics, optimizer=opt_factory)
            >>> print(_factory_to_dict(module.optimizer_no_decay_factory))
            {'_target_': 'torch.optim.adam.Adam', 'weight_decay': 0.0}
        """

        new_factory = copy.deepcopy(self.optimizer_factory)
        if "weight_decay" in new_factory.keywords:
            new_factory.keywords["weight_decay"] = 0.0
        else:
            logger.warning("No weight decay parameter found in optimizer factory. No changes made.")

        return new_factory

    def configure_optimizers(self):
        if self.optimizer_factory is None:
            raise ValueError("Optimizer factory is not set. Cannot configure optimizers.")

        params = [
            {"params": self._non_norm_bias_params(), "weight_decay": self.weight_decay},
            {"params": self._norm_bias_params(), "weight_decay": 0.0},
        ]

        optimizer = self.optimizer_no_decay_factory(params)

        if self.LR_scheduler_factory is None:
            return optimizer

        scheduler = self.LR_scheduler_factory(optimizer)

        LR_config = {
            "scheduler": scheduler,
            "frequency": 1,
        }

        if isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
            # ReduceLROnPlateau requires observing stable trends to make a conclusion about LR decay, so an
            # epcoh level interval is more appropriate.

            LR_config["monitor"] = "tuning/loss"
            LR_config["strict"] = True
            LR_config["interval"] = "epoch"
        else:
            # All other schedulers operate at a step level as they do not monitor the loss to make a
            # conclusion about LR decay.

            LR_config["interval"] = "step"

        return {"optimizer": optimizer, "lr_scheduler": LR_config}

    def predict_step(self, batch: MEDSTorchBatch):
        """Produces generated trajectories for a given batch of data."""
        return self.model.generate(batch)



import logging
import shutil
from datetime import UTC, datetime
from importlib.resources import files
from pathlib import Path

import hydra
from hydra.utils import instantiate
from lightning.pytorch import seed_everything
from meds import held_out_split, train_split, tuning_split
from meds_torchdata import MEDSTorchDataConfig
from MEDS_transforms.runner import load_yaml_file  # noqa: F401
from omegaconf import DictConfig, OmegaConf

from .generation import format_trajectories, get_timeline_end_token_idx
from .training import MEICARModule
from .utils import hash_based_seed, int_prod, resolve_generation_context_size  # noqa: F401

logger = logging.getLogger(__name__)

CONFIGS = files("MEDS_EIC_AR") / "configs"

MEDSTorchDataConfig.add_to_config_store("datamodule/config")


@hydra.main(version_base=None, config_path=str(CONFIGS), config_name="_pretrain")
def pretrain(cfg: DictConfig):
    st = datetime.now(tz=UTC)

    OmegaConf.save(cfg, Path(cfg.output_dir) / "config.yaml")

    D = instantiate(cfg.datamodule)

    gpt_kwargs = {"vocab_size": D.config.vocab_size, "eos_token_id": get_timeline_end_token_idx(D.config)}

    M = instantiate(
        cfg.lightning_module,
        model={"gpt_kwargs": gpt_kwargs},
        metrics={"vocab_size": D.config.vocab_size},
    )

    if M.model.do_demo or cfg.get("seed", None):
        seed_everything(cfg.get("seed", 1), workers=True)

    trainer = instantiate(cfg.trainer)

    trainer.fit(model=M, datamodule=D)

    best_ckpt_path = Path(trainer.checkpoint_callback.best_model_path)
    if not best_ckpt_path.is_file():
        raise ValueError("No best checkpoint reported.")

    output_fp = Path(cfg.output_dir) / "best_model.ckpt"
    shutil.copyfile(best_ckpt_path, output_fp)

    best_score = trainer.checkpoint_callback.best_model_score

    logger.info(f"Best checkpoint (with score {best_score:.2f}) copied to {output_fp!s}.")
    logger.info(f"Training complete in {datetime.now(tz=UTC) - st}")


@hydra.main(version_base=None, config_path=str(CONFIGS), config_name="_generate_trajectories")
def generate_trajectories(cfg: DictConfig):
    st = datetime.now(tz=UTC)

    D = instantiate(cfg.datamodule)

    M = MEICARModule.load_from_checkpoint(Path(cfg.ckpt_path))
    M.eval()

    trainer = instantiate(cfg.trainer)

    inference = cfg.inference

    if cfg.get("seed", None):
        seed_everything(cfg.get("seed", 1), workers=True)

    for split in inference.generate_for_splits:
        if split == train_split:
            dataloader = D.train_dataloader()
        elif split == tuning_split:
            dataloader = D.val_dataloader()
        elif split == held_out_split:
            dataloader = D.test_dataloader()
        else:
            raise ValueError(f"Unknown split {split}.")

        for sample in range(inference.N_trajectories_per_task_sample):
            out_fp = Path(cfg.output_dir) / split / f"{sample}.parquet"
            out_fp.parent.mkdir(parents=True, exist_ok=True)

            if out_fp.is_file() and not cfg.do_overwrite:
                logger.info(f"Skipping {out_fp} as it already exists.")
                continue
            else:
                out_fp.parent.mkdir(parents=True, exist_ok=True)

            seed = hash_based_seed(cfg.get("seed", None), split, sample)

            logger.info(f"Generating trajectories for {split} sample {sample} to {out_fp} with seed {seed}.")

            seed_everything(seed, workers=True)
            predictions = trainer.predict(model=M, dataloaders=dataloader)
            predictions_df = format_trajectories(dataloader.dataset, predictions)
            predictions_df.write_parquet(out_fp, use_pyarrow=True)

    logger.info(f"Generation of trajectories complete in {datetime.now(tz=UTC) - st}")



from datetime import timedelta
from typing import NamedTuple

import polars as pl
import torch
from meds import DataSchema, LabelSchema
from meds_torchdata import MEDSPytorchDataset
from MEDS_transforms.stages.add_time_derived_measurements.utils import normalize_time_unit

TIMELINE_DELTA_TOKEN = "TIMELINE//DELTA"
TASK_SAMPLE_ID_COL = "_task_sample_id"


class CodeInformation(NamedTuple):
    code: str
    value_prob: float
    value_mean: float | None


def get_code_information(dataset: MEDSPytorchDataset) -> dict[int, CodeInformation]:
    """Returns a dictionary mapping code indices to their code strings and numeric value means.

    Args:
        dataset: The dataset used for generation.

    Returns:
        A dictionary mapping code indices to their code strings and numeric value means.

    Examples:
        >>> get_code_information(pytorch_dataset)
        {1: CodeInformation(code='ADMISSION//CARDIAC', value_prob=0.0, value_mean=None),
         2: CodeInformation(code='ADMISSION//ORTHOPEDIC', value_prob=0.0, value_mean=None),
         3: CodeInformation(code='ADMISSION//PULMONARY', value_prob=0.0, value_mean=None),
         4: CodeInformation(code='DISCHARGE', value_prob=0.0, value_mean=None),
         5: CodeInformation(code='DOB', value_prob=0.0, value_mean=None),
         6: CodeInformation(code='EYE_COLOR//BLUE', value_prob=0.0, value_mean=None),
         7: CodeInformation(code='EYE_COLOR//BROWN', value_prob=0.0, value_mean=None),
         8: CodeInformation(code='EYE_COLOR//HAZEL', value_prob=0.0, value_mean=None),
         9: CodeInformation(code='HEIGHT//value_[156.4856,160.39531)', value_prob=1.0, value_mean=156.4...),
         10: CodeInformation(code='HEIGHT//value_[160.39531,164.68689)', value_prob=1.0, value_mean=160.3...),
         11: CodeInformation(code='HEIGHT//value_[164.68689,175.27112)', value_prob=1.0, value_mean=164.6...),
         12: CodeInformation(code='HEIGHT//value_[175.27112,inf)', value_prob=1.0, value_mean=175.2...),
         13: CodeInformation(code='HR//value_[-inf,102.6)', value_prob=1.0, value_mean=86.0),
         14: CodeInformation(code='HR//value_[102.6,105.1)', value_prob=1.0, value_mean=102.5999984741211),
         15: CodeInformation(code='HR//value_[105.1,107.5)', value_prob=1.0, value_mean=105.0999984741211),
         16: CodeInformation(code='HR//value_[107.5,107.7)', value_prob=1.0, value_mean=107.5),
         17: CodeInformation(code='HR//value_[107.7,112.5)', value_prob=1.0, value_mean=108.3499984741211),
         18: CodeInformation(code='HR//value_[112.5,112.6)', value_prob=1.0, value_mean=112.5),
         19: CodeInformation(code='HR//value_[112.6,113.4)', value_prob=1.0, value_mean=112.5999984741211),
         20: CodeInformation(code='HR//value_[113.4,114.1)', value_prob=1.0, value_mean=113.4000015258789),
         21: CodeInformation(code='HR//value_[114.1,119.8)', value_prob=1.0, value_mean=114.0999984741211),
         22: CodeInformation(code='HR//value_[119.8,inf)', value_prob=1.0, value_mean=145.0),
         23: CodeInformation(code='TEMP//value_[-inf,95.8)', value_prob=1.0, value_mean=95.5),
         24: CodeInformation(code='TEMP//value_[100.0,100.1)', value_prob=1.0, value_mean=100.0),
         25: CodeInformation(code='TEMP//value_[100.1,inf)', value_prob=1.0, value_mean=100.25),
         26: CodeInformation(code='TEMP//value_[95.8,96.0)', value_prob=1.0, value_mean=95.80000305175781),
         27: CodeInformation(code='TEMP//value_[96.0,96.2)', value_prob=1.0, value_mean=96.0),
         28: CodeInformation(code='TEMP//value_[96.2,97.8)', value_prob=1.0, value_mean=96.19999694824219),
         29: CodeInformation(code='TEMP//value_[97.8,99.9)', value_prob=1.0, value_mean=98.80000305175781),
         30: CodeInformation(code='TEMP//value_[99.9,100.0)', value_prob=1.0, value_mean=99.9000015258789),
         31: CodeInformation(code='TIMELINE//DELTA//years//value_...', value_prob=1.0, value_mean=3...e-06),
         32: CodeInformation(code='TIMELINE//DELTA//years//value_...', value_prob=1.0, value_mean=1...e-05),
         33: CodeInformation(code='TIMELINE//DELTA//years//value_...', value_prob=1.0, value_mean=4...e-05),
         34: CodeInformation(code='TIMELINE//DELTA//years//value_...', value_prob=1.0, value_mean=6...e-05),
         35: CodeInformation(code='TIMELINE//DELTA//years//value_...', value_prob=1.0, value_mean=0...),
         36: CodeInformation(code='TIMELINE//DELTA//years//value_...', value_prob=1.0, value_mean=31...),
         37: CodeInformation(code='TIMELINE//END', value_prob=0.0, value_mean=None),
         38: CodeInformation(code='TIMELINE//START', value_prob=0.0, value_mean=None)}
    """
    code_information = {}

    columns = ["code", "code/vocab_index", "code/n_occurrences", "values/n_occurrences", "values/sum"]
    code_metadata_df = pl.read_parquet(dataset.config.code_metadata_fp, columns=columns, use_pyarrow=True)

    for row in code_metadata_df.to_dicts():
        has_value_prob = row["values/n_occurrences"] / row["code/n_occurrences"]
        value_mean = (row["values/sum"] / row["values/n_occurrences"]) if has_value_prob else None
        code_information[row["code/vocab_index"]] = CodeInformation(
            code=row["code"],
            value_prob=has_value_prob,
            value_mean=value_mean,
        )

    return code_information


def format_trajectory_batch(
    schema_chunk: pl.DataFrame,
    generated_code_indices: torch.LongTensor,
    code_information: dict[int, CodeInformation],
) -> pl.DataFrame:
    """Formats a single batch of generated outputs into a MEDS-like dataframe format.

    Args:
        schema_chunk: The chunk of the dataset's schema dataframe corresponding to this generated batch. This
            dataframe must have the following columns:
              - `"subject_id"`: The subject ID of the patient.
              - `"prediction_time"`: The time after which no data can be ingested for this prediction. This
                is, implicitly, assumed to be the start time of the generated window. TODO(mmd): This
                assumption may not always be valid, and we don't have an exposed endpoint for the true time of
                the last event in the input!
              - `"task_sample_id"`: The task sample ID of the patient. This is a unique identifier for the
                task sample, which is useful as there may be different task samples for the same patient, and
                we don't wish our generated trajectories to intersect.
        generated_code_indices: The generated codes for this batch.
        code_information: The code information mapping from code indices to their string representations and
            numeric value means.

    Returns:
        A polars dataframe containing this batch of generated trajectory data in a MEDS-like format.

    Examples:
        >>> schema_df = pl.DataFrame({
        ...     "_task_sample_id": [1, 2, 3, 4, 5],
        ...     "subject_id": [1, 2, 3, 1, 4],
        ...     "prediction_time": [
        ...         datetime(1993, 1, 1),
        ...         datetime(2000, 1, 2),
        ...         datetime(1973, 1, 3),
        ...         datetime(2002, 10, 12),
        ...         datetime(2002, 10, 12),
        ...     ],
        ... })
        >>> generated_code_indices = torch.LongTensor([
        ...     [2, 4, 5, 4, 3, 4, 5, 5, 1, 6],
        ...     [1, 7, 9, 0, 0, 0, 0, 0, 0, 0],
        ...     [4, 7, 3, 6, 1, 9, 0, 0, 0, 0],
        ...     [6, 1, 8, 9, 0, 0, 0, 0, 0, 0],
        ...     [9, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ... ])
        >>> code_information = {
        ...     1: CodeInformation(code='TIMELINE//DELTA//years//value_A', value_prob=1.0, value_mean=1.0),
        ...     2: CodeInformation(code='TIMELINE//DELTA//years//value_A', value_prob=1.0, value_mean=0.1),
        ...     3: CodeInformation(code='TIMELINE//DELTA//years//value_A', value_prob=1.0, value_mean=0.001),
        ...     4: CodeInformation(code='HR', value_prob=0.0, value_mean=None),
        ...     5: CodeInformation(code='TEMP//A', value_prob=1.0, value_mean=99.0),
        ...     6: CodeInformation(code='TEMP//B', value_prob=1.0, value_mean=101.0),
        ...     7: CodeInformation(code='TEMP//C', value_prob=1.0, value_mean=96.8),
        ...     8: CodeInformation(code='DX//1', value_prob=0.0, value_mean=None),
        ...     9: CodeInformation(code='TIMELINE//END', value_prob=0.0, value_mean=None)
        ... }

    With this setup, we are generating the following strings of events for these patients:

        >>> schema_df
        shape: (5, 3)
        ┌─────────────────┬────────────┬─────────────────────┐
        │ _task_sample_id ┆ subject_id ┆ prediction_time     │
        │ ---             ┆ ---        ┆ ---                 │
        │ i64             ┆ i64        ┆ datetime[μs]        │
        ╞═════════════════╪════════════╪═════════════════════╡
        │ 1               ┆ 1          ┆ 1993-01-01 00:00:00 │
        │ 2               ┆ 2          ┆ 2000-01-02 00:00:00 │
        │ 3               ┆ 3          ┆ 1973-01-03 00:00:00 │
        │ 4               ┆ 1          ┆ 2002-10-12 00:00:00 │
        │ 5               ┆ 4          ┆ 2002-10-12 00:00:00 │
        └─────────────────┴────────────┴─────────────────────┘

    For task sample 1 (starting at 1993-01-01), we generate:
      - (2) A timeline delta of 0.1 years, which is 35.6 days. Event time: ~1993-02-06, 12:30
      - (4) A HR event, with no numeric value.
      - (5) A TEMP//A event, with a numeric value of 99.0.
      - (4) A HR event, with no numeric value.
      - (3) A timeline delta of 0.001 years, which is 0.0365 days (8.7 hours). Event time: ~1993-02-06, 21:20
      - (4) A HR event, with no numeric value.
      - (5) A TEMP//A event, with a numeric value of 99.0.
      - (5) A TEMP//A event, with a numeric value of 99.0.
      - (1) A timeline delta of 1.0 year. Event time: ~1994-02-06, 21:20
      - (6) A TEMP//B event, with a numeric value of 101.0.
    For task sample 2 (starting at 2000-01-02), we generate:
      - (1) A timeline delta of 1.0 year. Event time: ~2001-01-02
      - (7) A TEMP//C event, with a numeric value of 96.8.
      - (9) A TIMELINE//END event, with no numeric value.
    For task sample 3 (starting at 1973-01-03), we generate the following:
      - (4) A HR event, with no numeric value.
      - (7) A TEMP//C event, with a numeric value of 96.8.
      - (3) A timeline delta of 0.001 years, which is 0.0365 days (8.7 hours). Event time: ~1973-01-03, 08:45
      - (6) A TEMP//B event, with a numeric value of 101.0.
      - (1) A timeline delta of 1.0 year. Event time: ~1974-01-03, 08:45
      - (9) A TIMELINE//END event, with no numeric value.
    For task sample 4 (starting at 2002-10-12), we generate:
      - (6) A TEMP//B event, with a numeric value of 101.0.
      - (1) A timeline delta of 1.0 year. Event time: ~2003-10-12
      - (8) A DX//1 event, with no numeric value.
      - (9) A TIMELINE//END event, with no numeric value.
    For task sample 5 (starting at 2002-10-12), we generate:
      - (9) A TIMELINE//END event, with no numeric value.

        >>> _ = pl.Config().set_tbl_rows(-1)
        >>> format_trajectory_batch(schema_df, generated_code_indices, code_information)
        shape: (24, 5)
        ┌─────────────────┬────────────┬─────────────────────────┬─────────────────────────┬───────────────┐
        │ _task_sample_id ┆ subject_id ┆ time                    ┆ code                    ┆ numeric_value │
        │ ---             ┆ ---        ┆ ---                     ┆ ---                     ┆ ---           │
        │ i64             ┆ i64        ┆ datetime[μs]            ┆ str                     ┆ f32           │
        ╞═════════════════╪════════════╪═════════════════════════╪═════════════════════════╪═══════════════╡
        │ 1               ┆ 1          ┆ 1993-02-06 12:34:52.608 ┆ TIMELINE//DELTA//years/ ┆ 0.1           │
        │                 ┆            ┆                         ┆ /value_…                ┆               │
        │ 1               ┆ 1          ┆ 1993-02-06 12:34:52.608 ┆ HR                      ┆ null          │
        │ 1               ┆ 1          ┆ 1993-02-06 12:34:52.608 ┆ TEMP//A                 ┆ 99.0          │
        │ 1               ┆ 1          ┆ 1993-02-06 12:34:52.608 ┆ HR                      ┆ null          │
        │ 1               ┆ 1          ┆ 1993-02-06              ┆ TIMELINE//DELTA//years/ ┆ 0.001         │
        │                 ┆            ┆ 21:20:49.534080         ┆ /value_…                ┆               │
        │ 1               ┆ 1          ┆ 1993-02-06              ┆ HR                      ┆ null          │
        │                 ┆            ┆ 21:20:49.534080         ┆                         ┆               │
        │ 1               ┆ 1          ┆ 1993-02-06              ┆ TEMP//A                 ┆ 99.0          │
        │                 ┆            ┆ 21:20:49.534080         ┆                         ┆               │
        │ 1               ┆ 1          ┆ 1993-02-06              ┆ TEMP//A                 ┆ 99.0          │
        │                 ┆            ┆ 21:20:49.534080         ┆                         ┆               │
        │ 1               ┆ 1          ┆ 1994-02-07              ┆ TIMELINE//DELTA//years/ ┆ 1.0           │
        │                 ┆            ┆ 03:09:35.614080         ┆ /value_…                ┆               │
        │ 1               ┆ 1          ┆ 1994-02-07              ┆ TEMP//B                 ┆ 101.0         │
        │                 ┆            ┆ 03:09:35.614080         ┆                         ┆               │
        │ 2               ┆ 2          ┆ 2001-01-01 05:48:46.080 ┆ TIMELINE//DELTA//years/ ┆ 1.0           │
        │                 ┆            ┆                         ┆ /value_…                ┆               │
        │ 2               ┆ 2          ┆ 2001-01-01 05:48:46.080 ┆ TEMP//C                 ┆ 96.800003     │
        │ 2               ┆ 2          ┆ 2001-01-01 05:48:46.080 ┆ TIMELINE//END           ┆ null          │
        │ 3               ┆ 3          ┆ 1973-01-03 00:00:00     ┆ HR                      ┆ null          │
        │ 3               ┆ 3          ┆ 1973-01-03 00:00:00     ┆ TEMP//C                 ┆ 96.800003     │
        │ 3               ┆ 3          ┆ 1973-01-03              ┆ TIMELINE//DELTA//years/ ┆ 0.001         │
        │                 ┆            ┆ 08:45:56.926080         ┆ /value_…                ┆               │
        │ 3               ┆ 3          ┆ 1973-01-03              ┆ TEMP//B                 ┆ 101.0         │
        │                 ┆            ┆ 08:45:56.926080         ┆                         ┆               │
        │ 3               ┆ 3          ┆ 1974-01-03              ┆ TIMELINE//DELTA//years/ ┆ 1.0           │
        │                 ┆            ┆ 14:34:43.006080         ┆ /value_…                ┆               │
        │ 3               ┆ 3          ┆ 1974-01-03              ┆ TIMELINE//END           ┆ null          │
        │                 ┆            ┆ 14:34:43.006080         ┆                         ┆               │
        │ 4               ┆ 1          ┆ 2002-10-12 00:00:00     ┆ TEMP//B                 ┆ 101.0         │
        │ 4               ┆ 1          ┆ 2003-10-12 05:48:46.080 ┆ TIMELINE//DELTA//years/ ┆ 1.0           │
        │                 ┆            ┆                         ┆ /value_…                ┆               │
        │ 4               ┆ 1          ┆ 2003-10-12 05:48:46.080 ┆ DX//1                   ┆ null          │
        │ 4               ┆ 1          ┆ 2003-10-12 05:48:46.080 ┆ TIMELINE//END           ┆ null          │
        │ 5               ┆ 4          ┆ 2002-10-12 00:00:00     ┆ TIMELINE//END           ┆ null          │
        └─────────────────┴────────────┴─────────────────────────┴─────────────────────────┴───────────────┘

    This function is robust to the unit string used for the timeline delta tokens, provided it conforms to the
    set recognized in
    [`MEDS_transforms.stages.add_time_derived_measurements.utils.normalize_time_unit`](https://meds-transforms.readthedocs.io/en/latest/api/MEDS_transforms/stages/add_time_derived_measurements/utils/#MEDS_transforms.stages.add_time_derived_measurements.utils.normalize_time_unit).
    However, note that as time deltas are based on aggregate units (not calendar units), they won't
    necessarily universally correspond to true calendar months or years.

        >>> code_information = {
        ...     1: CodeInformation(code='TIMELINE//DELTA//s//A', value_prob=1.0, value_mean=1.0),
        ...     2: CodeInformation(code='TIMELINE//DELTA//days//A', value_prob=1.0, value_mean=1.0),
        ...     3: CodeInformation(code='TIMELINE//DELTA//wks//B', value_prob=1.0, value_mean=1.0),
        ...     4: CodeInformation(code='TIMELINE//DELTA//mos//C', value_prob=1.0, value_mean=1.0),
        ...     5: CodeInformation(code='TIMELINE//DELTA//yrs//C', value_prob=1.0, value_mean=1.0),
        ... }
        >>> schema_df = pl.DataFrame({
        ...     "_task_sample_id": [1], "subject_id": [1], "prediction_time": [datetime(1993, 1, 1)],
        ... })
        >>> generated_code_indices = torch.LongTensor([[1, 2, 3, 4, 5]])
        >>> format_trajectory_batch(schema_df, generated_code_indices, code_information)
        shape: (5, 5)
        ┌─────────────────┬────────────┬─────────────────────────┬─────────────────────────┬───────────────┐
        │ _task_sample_id ┆ subject_id ┆ time                    ┆ code                    ┆ numeric_value │
        │ ---             ┆ ---        ┆ ---                     ┆ ---                     ┆ ---           │
        │ i64             ┆ i64        ┆ datetime[μs]            ┆ str                     ┆ f32           │
        ╞═════════════════╪════════════╪═════════════════════════╪═════════════════════════╪═══════════════╡
        │ 1               ┆ 1          ┆ 1993-01-01 00:00:01     ┆ TIMELINE//DELTA//s//A   ┆ 1.0           │
        │ 1               ┆ 1          ┆ 1993-01-02 00:00:01     ┆ TIMELINE//DELTA//days// ┆ 1.0           │
        │                 ┆            ┆                         ┆ A                       ┆               │
        │ 1               ┆ 1          ┆ 1993-01-09 00:00:01     ┆ TIMELINE//DELTA//wks//B ┆ 1.0           │
        │ 1               ┆ 1          ┆ 1993-02-08 10:29:07     ┆ TIMELINE//DELTA//mos//C ┆ 1.0           │
        │ 1               ┆ 1          ┆ 1994-02-08 16:17:53.080 ┆ TIMELINE//DELTA//yrs//C ┆ 1.0           │
        └─────────────────┴────────────┴─────────────────────────┴─────────────────────────┴───────────────┘

    Note than an error is raised if the schema chunk does not have the same batch size as the generated
    tokens:

        >>> format_trajectory_batch(schema_df[:0], generated_code_indices, code_information)
        Traceback (most recent call last):
            ...
        ValueError: Batch size 1 does not match schema chunk size 0
    """

    batch_size = generated_code_indices.shape[0]
    if batch_size != schema_chunk.shape[0]:
        raise ValueError(f"Batch size {batch_size} does not match schema chunk size {schema_chunk.shape[0]}")

    rows = []
    for i in range(batch_size):
        subject_id = schema_chunk.select(DataSchema.subject_id_name)[i].item()
        time = schema_chunk.select(LabelSchema.prediction_time_name)[i].item()
        task_sample_id = schema_chunk.select(TASK_SAMPLE_ID_COL)[i].item()

        for code_idx in generated_code_indices[i]:
            if code_idx == 0:
                continue

            code_info = code_information[code_idx.item()]
            code = code_info.code
            value_mean = code_info.value_mean

            if code.startswith(TIMELINE_DELTA_TOKEN):
                unit = code.split("//")[-2]
                _, seconds_in_unit = normalize_time_unit(unit)
                seconds = value_mean * seconds_in_unit
                time += timedelta(seconds=seconds)

            rows.append(
                {
                    TASK_SAMPLE_ID_COL: task_sample_id,
                    DataSchema.subject_id_name: subject_id,
                    DataSchema.time_name: time,
                    DataSchema.code_name: code,
                    DataSchema.numeric_value_name: value_mean,
                }
            )

    return pl.DataFrame(
        rows,
        schema={
            TASK_SAMPLE_ID_COL: pl.Int64,
            DataSchema.subject_id_name: pl.Int64,
            DataSchema.time_name: pl.Datetime,
            DataSchema.code_name: pl.Utf8,
            DataSchema.numeric_value_name: pl.Float32,
        },
    )


def format_trajectories(
    dataset: MEDSPytorchDataset,
    generated_outputs: list[torch.LongTensor],
) -> pl.DataFrame:
    """Transfomrs the generated outputs into a MEDS-like dataframe format of continued trajectories.

    Args:
        dataset: The dataset used for generation.
        generated_outputs: The generated outputs. This is formatted as a list of generated samples that should
            be of the same length as the dataframe.

    Returns:
        A polars dataframe containing the generated trajectories in a MEDS-like format.

    Raises:
        ValueError: If the passed dataset does not yield code info with values strictly either always
            occurring or never occurring.

    Examples:
        >>> generated_code_indices = [
        ...     torch.LongTensor([[31, 4, 15, 4, 3, 4, 15, 15, 33, 16], [32, 17, 33, 16, 1, 37, 0, 0, 0, 0]]),
        ...     torch.LongTensor([[36, 17, 37], [37, 0, 0]]),
        ... ]
        >>> _ = pl.Config().set_tbl_rows(-1)
        >>> format_trajectories(pytorch_dataset_with_task, generated_code_indices)
        shape: (20, 5)
        ┌─────────────────┬────────────┬─────────────────────┬─────────────────────────────┬───────────────┐
        │ _task_sample_id ┆ subject_id ┆ time                ┆ code                        ┆ numeric_value │
        │ ---             ┆ ---        ┆ ---                 ┆ ---                         ┆ ---           │
        │ i64             ┆ i64        ┆ datetime[μs]        ┆ str                         ┆ f32           │
        ╞═════════════════╪════════════╪═════════════════════╪═════════════════════════════╪═══════════════╡
        │ 0               ┆ 239684     ┆ 2010-05-11 18:01:40 ┆ TIMELINE//DELTA//years//val ┆ 0.000003      │
        │                 ┆            ┆                     ┆ ue_…                        ┆               │
        │ 0               ┆ 239684     ┆ 2010-05-11 18:01:40 ┆ DISCHARGE                   ┆ null          │
        │ 0               ┆ 239684     ┆ 2010-05-11 18:01:40 ┆ HR//value_[105.1,107.5)     ┆ 105.099998    │
        │ 0               ┆ 239684     ┆ 2010-05-11 18:01:40 ┆ DISCHARGE                   ┆ null          │
        │ 0               ┆ 239684     ┆ 2010-05-11 18:01:40 ┆ ADMISSION//PULMONARY        ┆ null          │
        │ 0               ┆ 239684     ┆ 2010-05-11 18:01:40 ┆ DISCHARGE                   ┆ null          │
        │ 0               ┆ 239684     ┆ 2010-05-11 18:01:40 ┆ HR//value_[105.1,107.5)     ┆ 105.099998    │
        │ 0               ┆ 239684     ┆ 2010-05-11 18:01:40 ┆ HR//value_[105.1,107.5)     ┆ 105.099998    │
        │ 0               ┆ 239684     ┆ 2010-05-11          ┆ TIMELINE//DELTA//years//val ┆ 0.00004       │
        │                 ┆            ┆ 18:22:52.400010     ┆ ue_…                        ┆               │
        │ 0               ┆ 239684     ┆ 2010-05-11          ┆ HR//value_[107.5,107.7)     ┆ 107.5         │
        │                 ┆            ┆ 18:22:52.400010     ┆                             ┆               │
        │ 1               ┆ 239684     ┆ 2010-05-11          ┆ TIMELINE//DELTA//years//val ┆ 0.000015      │
        │                 ┆            ┆ 18:37:43.999983     ┆ ue_…                        ┆               │
        │ 1               ┆ 239684     ┆ 2010-05-11          ┆ HR//value_[107.7,112.5)     ┆ 108.349998    │
        │                 ┆            ┆ 18:37:43.999983     ┆                             ┆               │
        │ 1               ┆ 239684     ┆ 2010-05-11          ┆ TIMELINE//DELTA//years//val ┆ 0.00004       │
        │                 ┆            ┆ 18:58:56.399993     ┆ ue_…                        ┆               │
        │ 1               ┆ 239684     ┆ 2010-05-11          ┆ HR//value_[107.5,107.7)     ┆ 107.5         │
        │                 ┆            ┆ 18:58:56.399993     ┆                             ┆               │
        │ 1               ┆ 239684     ┆ 2010-05-11          ┆ ADMISSION//CARDIAC          ┆ null          │
        │                 ┆            ┆ 18:58:56.399993     ┆                             ┆               │
        │ 1               ┆ 239684     ┆ 2010-05-11          ┆ TIMELINE//END               ┆ null          │
        │                 ┆            ┆ 18:58:56.399993     ┆                             ┆               │
        │ 2               ┆ 239684     ┆ 2042-03-22          ┆ TIMELINE//DELTA//years//val ┆ 31.861664     │
        │                 ┆            ┆ 00:22:49.901777     ┆ ue_…                        ┆               │
        │ 2               ┆ 239684     ┆ 2042-03-22          ┆ HR//value_[107.7,112.5)     ┆ 108.349998    │
        │                 ┆            ┆ 00:22:49.901777     ┆                             ┆               │
        │ 2               ┆ 239684     ┆ 2042-03-22          ┆ TIMELINE//END               ┆ null          │
        │                 ┆            ┆ 00:22:49.901777     ┆                             ┆               │
        │ 3               ┆ 1195293    ┆ 2010-06-20 19:30:00 ┆ TIMELINE//END               ┆ null          │
        └─────────────────┴────────────┴─────────────────────┴─────────────────────────────┴───────────────┘

    If the dataset yields invalid code information, an error will be thrown:

        >>> with patch("MEDS_EIC_AR.generation.format_trajectories.get_code_information") as mock:
        ...     mock.return_value = {1: CodeInformation(code='HR', value_prob=0.5, value_mean=106.0)}
        ...     format_trajectories("fake dataset", generated_code_indices)
        Traceback (most recent call last):
          ...
        ValueError: Code HR has a value probability of 0.5, which is not 0.0 or 1.0. This is not supported.
    """

    code_information = get_code_information(dataset)

    for code_info in code_information.values():
        if code_info.value_prob not in {0.0, 1.0}:
            raise ValueError(
                f"Code {code_info.code} has a value probability of {code_info.value_prob}, "
                "which is not 0.0 or 1.0. This is not supported."
            )

    output_schema = (
        dataset.schema_df.select(DataSchema.subject_id_name, LabelSchema.prediction_time_name).clone()
    ).with_row_index(TASK_SAMPLE_ID_COL)

    batches_as_df = []
    st_i = 0
    for generated_codes in generated_outputs:
        batch_size = generated_codes.shape[0]

        # Get the schema chunk for this batch
        schema_chunk = output_schema.slice(st_i, batch_size)
        st_i += batch_size

        # Format the generated codes into a MEDS-like dataframe
        batch_df = format_trajectory_batch(schema_chunk, generated_codes, code_information)
        batches_as_df.append(batch_df)

    return pl.concat(batches_as_df)
